"""
Thresholding
-----------------
"""

from oats.threshold.pot import POTThreshold
from oats.threshold.spot import SPOTThreshold
from oats.threshold.quantile import QuantileThreshold
from oats.threshold.jenks import JenksThreshold
